@@ -11,6 +11,7 @@ import django
 from django.conf import settings
 from django.core.exceptions import ImproperlyConfigured, ValidationError
 from django.core.management.color import no_style
+from django.core.paginator import Paginator
 from django.db import DEFAULT_DB_ALIAS, connections
 from django.db.models.fields.related import ForeignObjectRel
 from django.db.models.query import QuerySet
@@ -41,6 +42,7 @@ logger = logging.getLogger(__name__)
 logger.addHandler(logging.NullHandler())
 
 USE_TRANSACTIONS = getattr(settings, 'IMPORT_EXPORT_USE_TRANSACTIONS', True)
+CHUNK_SIZE = getattr(settings, 'IMPORT_EXPORT_CHUNK_SIZE', 1)
 
 
 def get_related_model(field):
@@ -122,6 +124,12 @@ class ResourceOptions:
     The default value is False.
     """
 
+    chunk_size = None
+    """
+    Controls the chunk_size argument of Queryset.iterator or, 
+    if prefetch_related is used, the per_page attribute of Paginator.
+    """
+    
     skip_diff = False
     """
     Controls whether or not an instance should be diffed following import.
@@ -133,6 +141,28 @@ class ResourceOptions:
     The default value is False.
     """
 
+    use_bulk = False
+    """
+    Controls whether import operations should be performed in bulk.
+    By default, an object's save() method is called for each row in a data set.
+    When bulk is enabled, objects are saved using bulk operations.
+    """
+
+    batch_size = 1000
+    """
+    The batch_size parameter controls how many objects are created in a single query.
+    The default is to create objects in batches of 1000.
+    See `bulk_create() <https://docs.djangoproject.com/en/dev/ref/models/querysets/#bulk-create>`_.
+    This parameter is only used if ``use_bulk`` is True.
+    """
+
+    force_init_instance = False
+    """
+    If True, this parameter will prevent imports from checking the database for existing instances.
+    Enabling this parameter is a performance enhancement if your import dataset is guaranteed to 
+    contain new instances. 
+    """
+
 
 class DeclarativeMetaclass(type):
 
@@ -213,6 +243,11 @@ class Resource(metaclass=DeclarativeMetaclass):
         # cls.fields.
         self.fields = deepcopy(self.fields)
 
+        # lists to hold model instances in memory when bulk operations are enabled
+        self.create_instances = list()
+        self.update_instances = list()
+        self.delete_instances = list()
+
     @classmethod
     def get_result_class(self):
         """
@@ -247,6 +282,12 @@ class Resource(metaclass=DeclarativeMetaclass):
         else:
             return self._meta.use_transactions
 
+    def get_chunk_size(self):
+        if self._meta.chunk_size is None:
+            return CHUNK_SIZE
+        else:
+            return self._meta.chunk_size
+
     def get_fields(self, **kwargs):
         """
         Returns fields sorted according to
@@ -289,11 +330,77 @@ class Resource(metaclass=DeclarativeMetaclass):
         """
         Either fetches an already existing instance or initializes a new one.
         """
-        instance = self.get_instance(instance_loader, row)
-        if instance:
-            return (instance, False)
-        else:
-            return (self.init_instance(row), True)
+        if not self._meta.force_init_instance:
+            instance = self.get_instance(instance_loader, row)
+            if instance:
+                return (instance, False)
+        return (self.init_instance(row), True)
+
+    def get_import_id_fields(self):
+        """
+        """
+        return self._meta.import_id_fields
+
+    def get_bulk_update_fields(self):
+        """
+        Returns the fields to be included in calls to bulk_update().
+        ``import_id_fields`` are removed because `id` fields cannot be supplied to bulk_update().
+        """
+        return [f for f in self.fields if f not in self._meta.import_id_fields]
+
+    def bulk_create(self, using_transactions, dry_run, raise_errors, batch_size=None):
+        """
+        Creates objects by calling ``bulk_create``.
+        """
+        try:
+            if len(self.create_instances) > 0:
+                if not using_transactions and dry_run:
+                    pass
+                else:
+                    self._meta.model.objects.bulk_create(self.create_instances, batch_size=batch_size)
+        except Exception as e:
+            logger.exception(e)
+            if raise_errors:
+                raise e
+        finally:
+            self.create_instances.clear()
+
+    def bulk_update(self, using_transactions, dry_run, raise_errors, batch_size=None):
+        """
+        Updates objects by calling ``bulk_update``.
+        """
+        try:
+            if len(self.update_instances) > 0:
+                if not using_transactions and dry_run:
+                    pass
+                else:
+                    self._meta.model.objects.bulk_update(self.update_instances, self.get_bulk_update_fields(),
+                                                         batch_size=batch_size)
+        except Exception as e:
+            logger.exception(e)
+            if raise_errors:
+                raise e
+        finally:
+            self.update_instances.clear()
+
+    def bulk_delete(self, using_transactions, dry_run, raise_errors):
+        """
+        Deletes objects by filtering on a list of instances to be deleted,
+        then calling ``delete()`` on the entire queryset.
+        """
+        try:
+            if len(self.delete_instances) > 0:
+                if not using_transactions and dry_run:
+                    pass
+                else:
+                    delete_ids = [o.pk for o in self.delete_instances]
+                    self._meta.model.objects.filter(pk__in=delete_ids).delete()
+        except Exception as e:
+            logger.exception(e)
+            if raise_errors:
+                raise e
+        finally:
+            self.delete_instances.clear()
 
     def validate_instance(self, instance, import_validation_errors=None, validate_unique=True):
         """
@@ -327,15 +434,20 @@ class Resource(metaclass=DeclarativeMetaclass):
         """
         Takes care of saving the object to the database.
 
-        Keep in mind that this is done by calling ``instance.save()``, so
-        objects are not created in bulk!
+        Objects can be created in bulk if ``use_bulk`` is enabled.
         """
         self.before_save_instance(instance, using_transactions, dry_run)
-        if not using_transactions and dry_run:
-            # we don't have transactions and we want to do a dry_run
-            pass
+        if self._meta.use_bulk:
+            if instance.pk:
+                self.update_instances.append(instance)
+            else:
+                self.create_instances.append(instance)
         else:
-            instance.save()
+            if not using_transactions and dry_run:
+                # we don't have transactions and we want to do a dry_run
+                pass
+            else:
+                instance.save()
         self.after_save_instance(instance, using_transactions, dry_run)
 
     def before_save_instance(self, instance, using_transactions, dry_run):
@@ -353,13 +465,17 @@ class Resource(metaclass=DeclarativeMetaclass):
     def delete_instance(self, instance, using_transactions=True, dry_run=False):
         """
         Calls :meth:`instance.delete` as long as ``dry_run`` is not set.
+        If ``use_bulk`` then instances are appended to a list for bulk import.
         """
         self.before_delete_instance(instance, dry_run)
-        if not using_transactions and dry_run:
-            # we don't have transactions and we want to do a dry_run
-            pass
+        if self._meta.use_bulk:
+            self.delete_instances.append(instance)
         else:
-            instance.delete()
+            if not using_transactions and dry_run:
+                # we don't have transactions and we want to do a dry_run
+                pass
+            else:
+                instance.delete()
         self.after_delete_instance(instance, dry_run)
 
     def before_delete_instance(self, instance, dry_run):
@@ -411,8 +527,9 @@ class Resource(metaclass=DeclarativeMetaclass):
         Model instance need to have a primary key value before
         a many-to-many relationship can be used.
         """
-        if not using_transactions and dry_run:
+        if (not using_transactions and dry_run) or self._meta.use_bulk:
             # we don't have transactions and we want to do a dry_run
+            # OR use_bulk is enabled (m2m operations are not supported for bulk operations)
             pass
         else:
             for field in self.get_import_fields():
@@ -481,25 +598,25 @@ class Resource(metaclass=DeclarativeMetaclass):
         """
         pass
 
-    def before_import_row(self, row, **kwargs):
+    def before_import_row(self, row, row_number=None, **kwargs):
         """
         Override to add additional logic. Does nothing by default.
         """
         pass
 
-    def after_import_row(self, row, row_result, **kwargs):
+    def after_import_row(self, row, row_result, row_number=None, **kwargs):
         """
         Override to add additional logic. Does nothing by default.
         """
         pass
 
-    def after_import_instance(self, instance, new, **kwargs):
+    def after_import_instance(self, instance, new, row_number=None, **kwargs):
         """
         Override to add additional logic. Does nothing by default.
         """
         pass
 
-    def import_row(self, row, instance_loader, using_transactions=True, dry_run=False, **kwargs):
+    def import_row(self, row, instance_loader, using_transactions=True, dry_run=False, raise_errors=False, **kwargs):
         """
         Imports data from ``tablib.Dataset``. Refer to :doc:`import_workflow`
         for a more complete description of the whole import process.
@@ -575,6 +692,18 @@ class Resource(metaclass=DeclarativeMetaclass):
                 logger.debug(e, exc_info=e)
             tb_info = traceback.format_exc()
             row_result.errors.append(self.get_error_result_class()(e, tb_info, row))
+
+        if self._meta.use_bulk:
+            # persist a batch of rows
+            # because this is a batch, any exceptions are logged and not associated
+            # with a specific row
+            if len(self.create_instances) == self._meta.batch_size:
+                self.bulk_create(using_transactions, dry_run, raise_errors, batch_size=self._meta.batch_size)
+            if len(self.update_instances) == self._meta.batch_size:
+                self.bulk_update(using_transactions, dry_run, raise_errors, batch_size=self._meta.batch_size)
+            if len(self.delete_instances) == self._meta.batch_size:
+                self.bulk_delete(using_transactions, dry_run, raise_errors)
+
         return row_result
 
     def import_data(self, dataset, dry_run=False, raise_errors=False,
@@ -609,6 +738,9 @@ class Resource(metaclass=DeclarativeMetaclass):
 
         using_transactions = (use_transactions or dry_run) and supports_transactions
 
+        if self._meta.batch_size is not None and (not isinstance(self._meta.batch_size, int) or self._meta.batch_size < 0):
+            raise ValueError("Batch size must be a positive integer")
+
         with atomic_if_using_transaction(using_transactions):
             return self.import_data_inner(dataset, dry_run, raise_errors, using_transactions, collect_failed_rows, **kwargs)
 
@@ -647,6 +779,8 @@ class Resource(metaclass=DeclarativeMetaclass):
                     instance_loader,
                     using_transactions=using_transactions,
                     dry_run=dry_run,
+                    row_number=i,
+                    raise_errors=raise_errors,
                     **kwargs
                 )
             result.increment_row_result_total(row_result)
@@ -666,6 +800,13 @@ class Resource(metaclass=DeclarativeMetaclass):
                     self._meta.report_skipped):
                 result.append_row_result(row_result)
 
+        if self._meta.use_bulk:
+            # bulk persist any instances which are still pending
+            with atomic_if_using_transaction(using_transactions):
+                self.bulk_create(using_transactions, dry_run, raise_errors)
+                self.bulk_update(using_transactions, dry_run, raise_errors)
+                self.bulk_delete(using_transactions, dry_run, raise_errors)
+
         try:
             with atomic_if_using_transaction(using_transactions):
                 self.after_import(dataset, result, using_transactions, dry_run, **kwargs)
@@ -726,6 +867,23 @@ class Resource(metaclass=DeclarativeMetaclass):
     def get_user_visible_fields(self):
         return self.get_fields()
 
+    def iter_queryset(self, queryset):
+        if not isinstance(queryset, QuerySet):
+            yield from queryset
+        elif queryset._prefetch_related_lookups:
+            # Django's queryset.iterator ignores prefetch_related which might result
+            # in an excessive amount of db calls. Therefore we use pagination
+            # as a work-around
+            if not queryset.query.order_by:
+                # Paginator() throws a warning if there is no sorting
+                # attached to the queryset
+                queryset = queryset.order_by('pk')
+            paginator = Paginator(queryset, self.get_chunk_size())
+            for index in range(paginator.num_pages):
+                yield from paginator.get_page(index + 1)
+        else:
+            yield from queryset.iterator(chunk_size=self.get_chunk_size())
+
     def export(self, queryset=None, *args, **kwargs):
         """
         Exports a resource.
@@ -738,13 +896,7 @@ class Resource(metaclass=DeclarativeMetaclass):
         headers = self.get_export_headers()
         data = tablib.Dataset(headers=headers)
 
-        if isinstance(queryset, QuerySet):
-            # Iterate without the queryset cache, to avoid wasting memory when
-            # exporting large datasets.
-            iterable = queryset.iterator()
-        else:
-            iterable = queryset
-        for obj in iterable:
+        for obj in self.iter_queryset(queryset):
             data.append(self.export_resource(obj))
 
         self.after_export(queryset, data, *args, **kwargs)
@@ -935,11 +1087,6 @@ class ModelResource(Resource, metaclass=ModelDeclarativeMetaclass):
         )
         return field
 
-    def get_import_id_fields(self):
-        """
-        """
-        return self._meta.import_id_fields
-
     def get_queryset(self):
         """
         Returns a queryset of all objects for this model. Override this if you
